# FIXME: copyright

import base64
import binascii
import errors
import logging
import os
import struct
import sys
import tempfile

import fedora.client.AccountSystem
import nss.error
import nss.io
import nss.nss
import nss.ssl
import urlgrabber.grabber

import double_tls
import settings
import utils
import version

 # Infrastructure

class BridgeError(Exception):
    pass

class BridgeConfiguration(utils.NSSConfiguration, utils.Configuration):

    default_config_file = 'bridge.conf'

    def _add_defaults(self, defaults):
        super(BridgeConfiguration, self)._add_defaults(defaults)
        defaults.update({'bridge-cert-nickname': 'rpmsigner-bridge-cert',
                         'client-listen-port': 44334,
                         'required-fas-group': '',
                         'server-listen-port': 44333})

    def _read_configuration(self, parser):
        super(BridgeConfiguration, self)._read_configuration(parser)
        self.bridge_cert_nickname = parser.get('bridge', 'bridge-cert-nickname')
        self.client_listen_port = parser.getint('bridge', 'client-listen-port')
        self.required_fas_group = parser.get('bridge', 'required-fas-group')
        if self.required_fas_group == '':
            self.required_fas_group = None
        else:
            self.fas_user_name = parser.get('bridge', 'fas-user-name')
            self.fas_password = parser.get('bridge', 'fas-password')
        self.server_listen_port = parser.getint('bridge', 'server-listen-port')

_u32_struct = struct.Struct('!I')

def create_listen_sock(config, port):
    sock = nss.ssl.SSLSocket()
    # FIXME? does this belong in a finished product?
    sock.set_socket_option(nss.io.PR_SockOpt_Reuseaddr, True)
    sock.set_ssl_option(nss.ssl.SSL_REQUEST_CERTIFICATE, True)
    sock.set_ssl_option(nss.ssl.SSL_REQUIRE_CERTIFICATE, True)
    cert = nss.nss.find_cert_from_nickname(config.bridge_cert_nickname)
    sock.config_secure_server(cert, nss.nss.find_key_by_any_cert(cert),
                              cert.find_kea_type())
    sock.bind(nss.io.NetworkAddress(nss.io.PR_IpAddrAny, port))
    sock.listen()
    return sock

class InvalidRequestError(Exception):
    '''The client's request was invalid.'''
    pass

class ForwardingError(Exception):
    '''An error was detected while forwarding or modifying the communication.'''
    pass

def copy_data(dest, src, bytes):
    '''Copy bytes bytes from file-like src to file-like dst.'''
    while bytes > 0:
        data = src.read(min(bytes, 4096))
        dest.write(data)
        bytes -= len(data)

 # Request verification

class Field(object):
    '''A field.'''

    def __init__(self, name, optional=False):
        self.name = name
        self.__optional = optional

    def validate(self, value):
        '''Validate field value value.

        value is None if field is not present.

        '''
        if not self.__optional and value is None:
            raise InvalidRequestError('Required field %s missing' % self.name)

class StringField(Field):
    '''A string field.'''

    def validate(self, value):
        super(StringField, self).validate(value)
        if value is not None and not utils.string_is_safe(value):
            raise InvalidRequestError('Field %s is not printable' % self.name)

class BoolField(Field):
    '''A bool field.'''

    def validate(self, value):
        super(BoolField, self).validate(value)
        if value is not None and (len(value) != _u32_struct.size or
                                  _u32_struct.unpack(value)[0] not in (0, 1)):
            raise InvalidRequestError('Field %s is not a boolean' % self.name)

class YYYYMMDDField(Field):
    '''A date field, using the yyyy-mm-dd format.'''

    def validate(self, value):
        super(YYYYMMDDField, self).validate(value)
        if value is not None and not utils.yyyy_mm_dd_is_valid(value):
            raise InvalidRequestError('Field %s is not a valid date' %
                                      self.name)

class RequestType(object):
    '''A supported request type.'''

    def __init__(self, fields, max_payload=0):
        self.__fields = fields + (StringField('user'), StringField('op'))
        self.__known_fields = set()
        for f in self.__fields:
            self.__known_fields.add(f.name)
        self.__max_payload = max_payload

    def validate(self, fields, payload_size):
        '''Validate fields.'''
        for key in fields.iterkeys():
            if key not in self.__known_fields:
                raise InvalidRequestError('Unexpected field %s' % repr(key))
        for f in self.__fields:
            f.validate(fields.get(f.name))
        if payload_size > self.__max_payload:
            raise InvalidRequestError('Payload too large')

    def forward_request_payload(self, server_buf, client_buf, payload_size,
                                unused_fields):
        '''Forward (optionally modify) payload from client_buf to server_buf.'''
        server_buf.write(_u32_struct.pack(payload_size))
        copy_data(server_buf, client_buf, payload_size)

    def forward_reply_payload(self, client_buf, server_buf, payload_size):
        '''Forward (optionally modify) payload from server_buf to client_buf.'''
        client_buf.write(_u32_struct.pack(payload_size))
        copy_data(client_buf, server_buf, payload_size)

    def close(self):
        '''Deinitialize any costly state.'''
        pass


class SignRpmRequestType(RequestType):
    '''A specialized handler for the 'sign-rpm' request.'''

    def __init__(self, *args, **kwargs):
        super(SignRpmRequestType, self).__init__(*args, **kwargs)
        self.__koji_session = None
        self.__koji_rpm_info = None
        self.__rpm = None

    def forward_request_payload(self, server_buf, client_buf, payload_size,
                                fields):
        '''Forward (optionally modify) payload from client_buf to server_buf.'''
        # Don't import koji before opening sockets!  The rpm Python module
        # calls NSS_NoDB_Init() during its initialization, which breaks our
        # attempts to initialize nss with our certificate database.
        import koji

        self.__request_fields = fields
        if payload_size != 0:
            return super(SignRpmRequestType, self). \
                forward_request_payload(server_buf, client_buf, payload_size,
                                        fields)
        session = self.__koji_get_session()
        rpm = self.__koji_get_rpm_info(session)
        try:
            build = session.getBuild(rpm['build_id'])
            if build is None:
                raise ForwardingError('RPM has no build')
        except (utils.KojiError, koji.GenericError), e:
            raise ForwardingError('Koji connection failed: %s' % e.message)
        url = '/'.join((self.__koji_config['pkgurl'], build['package_name'],
                        build['version'], build['release'],
                        koji.pathinfo.rpm(rpm)))

        src = urlgrabber.grabber.urlopen(url)
        try:
            try:
                payload_size = int(src.hdr['Content-Length'])
            except KeyError:
                raise ForwardingError('Content-Length not returned for %s' %
                                      url)
            server_buf.write(_u32_struct.pack(payload_size))
            copy_data(server_buf, src, payload_size)
        finally:
            src.close()

    def forward_reply_payload(self, client_buf, server_buf, payload_size):
        # Don't import koji before opening sockets!  The rpm Python module
        # calls NSS_NoDB_Init() during its initialization, which breaks our
        # attempts to initialize nss with our certificate database.
        import koji

        if self.__request_fields.get('import-signature') == _u32_struct.pack(1):
            (fd, tmp_path) = tempfile.mkstemp(text=False)
            tmp_file = None
            try:
                tmp_file = os.fdopen(fd, 'w+')
                copy_data(tmp_file, server_buf, payload_size)
                tmp_file.flush()
                sigkey = koji.get_header_fields(tmp_path, ('siggpg',))['siggpg']
                if sigkey is None:
                    raise ForwardingError('Missing signature')
                # FIXME? This is not actually a key ID, but it is what Koji
                # uses.
                sigkey = koji.hex_string(sigkey[13:17])
                sighdr = koji.rip_rpm_sighdr(tmp_path)
                sighdr_digest = binascii.b2a_hex(utils.md5_digest(sighdr))

                session = self.__koji_get_session()
                rpm = self.__koji_get_rpm_info(session)
                try:
                    sigs = session.queryRPMSigs(rpm_id=rpm['id'], sigkey=sigkey)
                    assert len(sigs) <= 1
                    if len(sigs) > 0 and sigs[0]['sighash'] != sighdr_digest:
                        raise ForwardingError('A different signature was '
                                              'already imported')
                    if len(sigs) == 0:
                        session.addRPMSig(rpm['id'],
                                          base64.encodestring(sighdr))
                except (utils.KojiError, koji.GenericError), e:
                    raise ForwardingError('Koji connection failed: %s'
                                          % e.message)

                if (self.__request_fields.get('return-data') ==
                    _u32_struct.pack(0)):
                    client_buf.write(_u32_struct.pack(0))
                else:
                    tmp_file.seek(0)
                    client_buf.write(_u32_struct.pack(payload_size))
                    copy_data(client_buf, tmp_file, payload_size)
            finally:
                if tmp_file is not None:
                    tmp_file.close()
                os.remove(tmp_path)
        elif self.__request_fields.get('return-data') != _u32_struct.pack(0):
            super(SignRpmRequestType, self).forward_reply_payload(client_buf,
                                                                  server_buf,
                                                                  payload_size)
        else:
            client_buf.write(_u32_struct.pack(0))

    def __koji_get_session(self):
        '''Return a koji session, creating it if necessary.

        Also set up self.__koji_config.

        '''
        # Don't import koji before opening sockets!  The rpm Python module
        # calls NSS_NoDB_Init() during its initialization, which breaks our
        # attempts to initialize nss with our certificate database.
        import koji

        if self.__koji_session is None:
            try:
                self.__koji_config = utils.koji_read_config()
                # self.__request_fields['user'] safety was verified by
                # self.validate
                # FIXME FIXME: authenticate should be True, but that needs a
                # koji account that is allowed to use proxyuser
                self.__koji_session = \
                    utils.koji_connect(self.__koji_config,
                                       authenticate=False,
                                       proxyuser=self.__request_fields['user'])
            except (utils.KojiError, koji.GenericError), e:
                raise ForwardingError('Koji connection failed: %s' % e.message)
        return self.__koji_session

    def __koji_get_rpm_info(self, session):
        '''Return information about a rpm specified by self.__request_fields.'''
        # Don't import koji before opening sockets!  The rpm Python module
        # calls NSS_NoDB_Init() during its initialization, which breaks our
        # attempts to initialize nss with our certificate database.
        import koji

        if self.__koji_rpm_info is None:
            try:
                d = {'name': self.__request_fields['rpm-name'],
                     'version': self.__request_fields['rpm-version'],
                     'release': self.__request_fields['rpm-release'],
                     'arch': self.__request_fields['rpm-arch']}
            except KeyError:
                raise InvalidRequestError('Incomplete RPM identification')
            # String safety in d was verified by self.validate
            try:
                self.__koji_rpm_info = session.getRPM(d)
            except (utils.KojiError, koji.GenericError), e:
                raise ForwardingError('Koji connection failed: %s' % e.message)
            if self.__koji_rpm_info is None:
                raise ForwardingError('RPM not found')
        return self.__koji_rpm_info

    def close(self):
        if self.__koji_session is not None:
            utils.koji_disconnect(self.__koji_session)
            self.__koji_session = None

RT = RequestType
SF = StringField
request_types = {
    'list-users': RT(()),
    'new-user': RT((SF('name'), BoolField('admin', optional=True))),
    'delete-user': RT((SF('name'),)),
    'user-info': RT((SF('name'),)),
    'modify-user': RT((SF('name'), BoolField('admin', optional=True,),
                       SF('new-name', optional=True))),
    'key-user-info': RT((SF('name'), SF('key'))),
    'modify-key-user': RT((SF('name'), SF('key'),
                           BoolField('key-admin', optional=True))),
    'list-keys': RT(()),
    'new-key': RT((SF('key'), SF('initial-key-admin', optional=True),
                   SF('name-real', optional=True),
                   SF('name-comment', optional=True),
                   SF('name-email', optional=True),
                   YYYYMMDDField('expire-date', optional=True))),
    'delete-key': RT((SF('key'),)),
    'modify-key': RT((SF('key'), SF('new-name', optional=True))),
    'list-key-users': RT((SF('key'),)),
    'grant-key-access': RT((SF('key'), SF('name'))),
    'revoke-key-access': RT((SF('key'), SF('name'))),
    'get-public-key': RT((SF('key'),)),
    'change-passphrase': RT((SF('key'),)),
    'sign-text': RT((SF('key'),), max_payload=1024*1024*1024),
    'sign-data': RT((SF('key'),), max_payload=1024*1024*1024),
    'sign-rpm': SignRpmRequestType((SF('key'), SF('rpm-name', optional=True),
                                    SF('rpm-epoch', optional=True),
                                    SF('rpm-version', optional=True),
                                    SF('rpm-release', optional=True),
                                    SF('rpm-arch', optional=True),
                                    Field('rpm-sigmd5', optional=True),
                                    BoolField('import-signature',
                                              optional=True),
                                    BoolField('return-data', optional=True)),
                                   max_payload=1024*1024*1024),
    }
del RT

 # Request handling

class StoringProxy(object):
    '''A proxy for a double_tls.OuterBuffer that stores read outer data.'''

    def __init__(self, buf):
        self.__buf = buf
        self.__stored = ''

    def stored_read(self, bytes):
        '''Read bytes bytes from the buffer and store the result.'''
        data = self.__buf.read(bytes)
        self.__stored += data
        return data

    def stored_data(self):
        '''Return the currently stored data.

        Each piece of data is returned only once in repeated calls to this
        method.

        '''
        res = self.__stored
        self.__stored = ''
        return res

def handle_connection(client_buf, server_buf):
    '''Handle a single connection.'''
    # FIXME: handle server's reporting of unknown protocol version - there
    # is no inner session
    client_proxy = StoringProxy(client_buf)
    buf = client_proxy.stored_read(_u32_struct.size)
    client_version = _u32_struct.unpack(buf)[0]
    if client_version != version.protocol_version:
        raise InvalidRequestError('Unknown protocol version %d' %
                                  client_version)
    try:
        fields = utils.read_fields(client_proxy.stored_read)
    except utils.InvalidFieldsError, e:
        raise InvalidRequestError(e.message)
    buf = client_buf.read(_u32_struct.size)
    payload_size = _u32_struct.unpack(buf)[0]

    s = ', '.join(('%s = %s' % (repr(key), repr(value))
                   for (key, value) in fields.iteritems()))
    logging.info('Request: %s', s)
    try:
        op = fields['op']
    except KeyError:
        raise InvalidRequestError('op field missing')
    try:
        rt = request_types[op]
    except KeyError:
        raise InvalidRequestError('Unknown op value')

    try:
        rt.validate(fields, payload_size)
        server_buf.write(client_proxy.stored_data())
        rt.forward_request_payload(server_buf, client_buf, payload_size, fields)

        double_tls.bridge_inner_stream(client_buf, server_buf)

        server_proxy = StoringProxy(server_buf)
        buf = server_proxy.stored_read(_u32_struct.size)
        try:
            fields = utils.read_fields(server_proxy.stored_read)
        except utils.InvalidFieldsError, e:
            raise InvalidRequestError(e.message)
        error_code = _u32_struct.unpack(buf)[0]
        if error_code != errors.OK:
            msg = fields.get('message')
            if msg is not None:
                logging.info('Request error: %s, %s',
                             errors.message(error_code), msg)
            else:
                logging.info('Request error: %s', errors.message(error_code))
        # print repr(fields)
        server_proxy.stored_read(64) # Ignore value
        client_buf.write(server_proxy.stored_data())

        buf = server_buf.read(_u32_struct.size)
        payload_size = _u32_struct.unpack(buf)[0]
        rt.forward_reply_payload(client_buf, server_buf, payload_size)
    finally:
        rt.close()

    buf = server_buf.read(64)
    client_buf.write(buf)

def fas_user_is_in_group(config, user_name, group_name):
    '''Return True if user_name is in group.'''
    try:
        logging.debug('Authenticating user in FAS')
        conn = fedora.client.AccountSystem(username=config.fas_user_name,
                                           password=config.fas_password)
        person = conn.person_by_username(user_name)
        if len(person) == 0: # Not found
            return False
        for group in person.approved_memberships:
            if str(group['name']) == group_name:
                return True
        return False
    except (fedora.client.FedoraClientError,
            fedora.client.FedoraServiceError), e:
        raise BridgeError('Error communicating with FAS: %s' % e.message)

def main():
    options = utils.get_daemon_options('A signing server bridge',
                                       '~/.rpmsigner/bridge.conf')
    d = {}
    if settings.log_dir is not None:
        d['filename'] = os.path.join(settings.log_dir, 'bridge.log')
    logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s',
                        level=utils.logging_level_from_options(options), **d)
    try:
        config = BridgeConfiguration(options.config_file)
    except utils.ConfigurationError, e:
        sys.exit(e.message)

    try:
        utils.nss_init(config)

        server_listen_sock = create_listen_sock(config,
                                                config.server_listen_port)
        client_listen_sock = create_listen_sock(config,
                                                config.client_listen_port)
    except nss.error.NSPRError, e:
        logging.error('NSPR error: %s' % repr(e))
        sys.exit(1)

    utils.create_pid_file('rpmsigner_bridge')
    try:
        while True:
            try:
                client_sock = None
                logging.debug('Waiting for the server to connect')
                (server_sock, _) = server_listen_sock.accept()
                # FIXME? authenticate the server
                try:
                    logging.debug('Waiting for the client to connect')
                    (client_sock, _) = client_listen_sock.accept()

                    client_sock.force_handshake()
                    cert = client_sock.get_peer_certificate()
                    assert cert is not None
                    user_name = cert.subject_common_name
                    logging.info('Client with CN %s connected', repr(user_name))
                    if (config.required_fas_group is not None and
                        not fas_user_is_in_group(config, user_name,
                                                 config.required_fas_group)):
                        raise InvalidRequestError('User %s not allowed to connect'
                                                  % repr(user_name))

                    client_buf = double_tls.OuterBuffer(client_sock)
                    server_buf = double_tls.OuterBuffer(server_sock)

                    handle_connection(client_buf, server_buf)
                finally:
                    if client_sock is not None:
                        client_sock.close()
                    server_sock.close()
            except InvalidRequestError, e:
                logging.warning('Invalid request: %s', e.message)
            except IOError, e:
                logging.info('I/O error: %s', repr(e))
            except EOFError, e:
                logging.info('Unexpected EOF: %s', repr(e))
            except nss.error.NSPRError, e:
                logging.info('NSPR I/O error: %s', repr(e))
            except BridgeError, e:
                logging.warning(e.message)
            logging.debug('Request handling finished')
    except KeyboardInterrupt:
        logging.error('...', exc_info=True)
        pass # Silence is golden
    except:
        logging.error('Unexpected exception', exc_info=True)
        sys.exit(1)
    finally:
        utils.delete_pid_file('rpmsigner_bridge')

if __name__ == '__main__':
    main()
